/*
 *  Hardwarenahe Funktionen fr LegoBot
 *
 *	02.03.2010 BS abgespeckt und angepasst fr Roboter-AG
 *  28.08.2010 BS erweitert um Servo-Antrieb
 *	25.09.2010 BS Funktionsnamen gendert, zweiten Hindernistaster hinzugefgt
 *  14.04.2012 BS LED versetzt fr Bootloader
 *
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

#include "hardware.h"


#define LED_OFF			PORTA |= 0x10	
#define LED_ON			PORTA &= ~0x10
#define LED_TOGGLE		PORTA ^= 0x10


volatile uchar 	timer1;
volatile uchar 	timer2;
volatile uchar 	timer3;
volatile uchar 	timer4;
volatile uchar 	timer5;
volatile uchar 	timer_loop;
volatile uchar 	adc_arr[SENSORS];

volatile uchar	led_state;

uchar	adc_channel;
uchar 	timer_prescaler;
uchar 	timer_blink;
uchar   motor_links = 0;
uchar   motor_rechts = 0;
uchar   led_disable = 0;

/*
 * Interne Funktion: LED tatschlich setzen bzw. Zustand wechseln
 */
static void refreshLED(void)
{
	switch (led_state)
	{
		case 0:	
			LED_OFF;
			break;

		case 1:	
			LED_ON;
			break;

		case 2:
			if (!timer_blink)
			{
				timer_blink = TIMER_MS(100);
				LED_TOGGLE;
			}
			break;
	}
}


/*
 *	Interrupt-Routine fr Timer0, Systemtakt
 */
ISR (SIG_OVERFLOW0)
{
#ifndef SERVO_ANTRIEB
	if (!(--timer_prescaler))
	{
		timer_prescaler = 249;
#endif

		if (timer1)
			timer1--;
		if (timer2)
			timer2--;
		if (timer3)
			timer3--;
		if (timer4)
			timer4--;
		if (timer5)
			timer5--;
		if (timer_loop)
			timer_loop--;
		if (timer_blink)
			timer_blink--;

		if ( !led_disable )
		{
			refreshLED();
		}
#ifndef SERVO_ANTRIEB
	}
#endif

#ifdef SERVO_ANTRIEB
	// hier hat der Prescaler eine andere Aufgabe: er zhlt, wann der PWM-Impuls wirklich
	// ausgegeben werden soll, denn das darf nur jedes 10. Mal sein
	if (!(--timer_prescaler))
	{
		timer_prescaler = 10;
		
		// freigeben, d.h. die obersten 4 Bits auf 1 fr normale Port-Funktion
		// Allerdings nur dann, wenn die Motoren laufen sollen
		if ( motor_links )
		{
			TCCR0A |= 0x30;
		}
		
		if ( motor_rechts )
		{
			TCCR0A |= 0xC0;
		}
	}
	else
	{
		// sperren, d.h. die obersten 4 Bits auf 0 fr normale Port-Funktion.
		// Da die Ausgangsregister auf 0 sind, wird low ausgegeben
		TCCR0A &= 0x0F;
	}
	
#endif	
}


/*
 *	Interrupt-Routine fr ADC
 */
ISR (SIG_ADC)
{
	adc_arr[adc_channel++] = ADCH;

	if (adc_channel == 2)
	{
		adc_channel = 0;
	}

	ADMUX = adc_channel;
	ADCSRA |= (1 <<	ADSC);
}




void warteMillisec(uint value)
{
	while (value > 500)
	{
		timer_loop = TIMER_MS(500);
		value = value - 500;
		while (timer_loop)
			;
	}

	timer_loop = TIMER_MS(value);
	while (timer_loop)
		;
}


void setzeMotorLinks(int value)
{
#ifdef SERVO_ANTRIEB
    motor_links = value;
	
	if (!value)
	{
		// aus
		OCR0B = 0;
	}
	else
	{
		if (value > 0)
			OCR0B = 255 - 250;
		else
			OCR0B = 255 - 125;
	}
#else
	OCR0B = value;
#endif
}


void setzeMotorRechts(int value)
{
#ifdef SERVO_ANTRIEB
    motor_rechts = value;

	if (!value)
	{
		// aus
		OCR0A = 0;
	}
	else
	{
		if (value < 0)
			OCR0A = 255 - 250;
		else
			OCR0A = 255 - 125;
	}
#else
	OCR0A = value;
#endif
}


void setzeMotoren(int links, int rechts)
{
	setzeMotorLinks(links);
	setzeMotorRechts(rechts);
}


void setzeServoLinks( uchar percent )
{
	OCR0B = 16 + 15 * (uint) percent / 100;
}


void setzeServoRechts( uchar percent )
{
	OCR0A = 16 + 15 * (uint) percent / 100;
}


void setzeTon(uint freq)
{
	if (freq == 0)
	{
		TCCR1B &= ~0x01;	 	// Timer ausschalten durch Wegnehmen der Clock Source
	}
	else
	{	
		TCCR1B |= 0x01;		// Timer einschalten auf Prescaler 1
		ICR1 = 4000000L / freq - 1;
	}
}



uchar HindernisLinks(void)
{
    return ((PINB & (1 << PINB1)) == 0);
}


uchar HindernisRechts(void)
{
    return ((PINB & (1 << PINB0)) == 0);
}

static uchar SignalUS(void)
{
    return ((PINA & (1 << PINA2)) != 0);
}


void setzeLED(uchar state)
{
	led_state = state;
	led_disable = 0;
	refreshLED();
}


uchar SensorLinks(void)
{
	return adc_arr[1];
}


uchar SensorRechts(void)
{
	return adc_arr[0];
}


/**
 * wartet bis die Funktion SignalUS() den angegebenen Wert zurckliefert oder eine
 * maximale Anzahl von Timer0-Ticks abgelaufen ist.
 */
uint warteSignalTimeout( uchar state, uint timeoutTicks )
{
	uint  ticks = 0;
	uchar lastT0;
	
	lastT0 = TCNT0;
	while ( ( SignalUS() != state ) && ( ticks <= timeoutTicks ) )
	{
	    uchar currT0 = TCNT0;

		if ( currT0 < lastT0 )
		{
			// berlauf!
			ticks = ticks + 256 - lastT0 + currT0;
		}
		else
		{
		    ticks = ticks + ( currT0 - lastT0 );
		}
		
		lastT0 = currT0;
	}

	return ticks;
} 


uint AbstandCm( void )
{
	uchar i;
	uint  ticks = 0;
	uint  res = 0;
	
	// warten bis Eingang auf 0 geht. Sollte er zwar schon sein, aber bei fehlendem Echo
	// kann der Ausgang noch 1 sein. Ganz am Anfang ist er anscheinend sogar dauerhaft auf 1
	// bis der erste Trigger kommt. Daher verwenden wir also einen Timeout. Da der Ausgang bei 
	// ausbleibendem Echo knapp 250ms lang high ist, nehmen wir die 250ms als Timeout.
	warteSignalTimeout( 0, 250 * ( 1000 / 8 ) );
	
	// Trigger-Impuls:
	PORTA |= (1 << PORTA3);

	// mindestens 10s warten
	for ( i = 0; i < 30; i++ )
	{	
		asm volatile( "nop" ); 
	}
	
	PORTA &= ~(1 << PORTA3);

	// warten bis der Ausgang high wird. Auch hier brauchen wir einen kleinen Timeout,
	// falls kein Sensor angeschlossen ist
	warteSignalTimeout( 1, 1 * ( 1000 / 8 ) );
	
	if ( SignalUS() )
	{
		// ok, kein Timeout
		
		// Timer-Ticks zhlen bis der Eingang auf 0 geht oder eine maximale Zeit von 100ms um ist.
		// Da wir den Prescaler auf 64 gestellt haben, kommt ein Timertick alle 8 s 
		ticks = warteSignalTimeout( 0, 100 * ( 1000 / 8 ) );
	
		// bei Timeout geben wir 0 zurck
		if ( !SignalUS() )
		{
			// Umrechnen in cm. Bei einer Geschwindigkeit von 340m/s bentigt der Schall 29s pro cm.
			// Da er aber hin und zurck laufen muss, sind es 58,8s pro cm Abstand.
			// Das heit, man muss die gemessene Zeit in Mikrosekunden durch 59 teilen, um zu den cm zu kommen.
			// Das ganze machen wir in long, damit es zu keinem berlauf kommt:
			res = (unsigned long) ticks * 8UL / 59UL;
		}
	}
	
	return res;
}


static void waitBitTime( void )
{
	// eine Bit-Zeit entspricht fast genau 13 Timer0-Ticks
	uchar ticks = 0;
	uchar lastT0 = TCNT0;
	while ( ticks < 13 )
	{
		uchar currT0 = TCNT0;
		
		// normalerweise wrde man einfach schreiben: sum = sum + ( curr - last) 
		// und selbst ein berlauf wrde funktionieren. Aber leider werden die Berechnungen
		// von char und uchar int Integer durchgefhrt und dann muss der berlauf gesondert behandelt werden:
		if ( currT0 < lastT0 )
		{
			// berlauf! 256 ist gltig, da in Integer, also 16 Bit gerechnet wird!
			ticks = ticks + 256 - lastT0 + currT0;
		}
		else
		{
		    ticks = ticks + ( currT0 - lastT0 );
		}
		
		lastT0 = currT0;
	}
} 


void sendeZeichen( uchar ch )
{
	led_disable = 1;
	
	// Startbit: 0
	LED_ON;
	waitBitTime();
	
	// alle 8 Bits, niederwertiges zuerst
	uchar i = 0;
	while ( i < 8 )
	{
		if ( ch & 0x01 )
		{
		    LED_OFF;
		}
		else
		{
		    LED_ON;
		}
		waitBitTime();
		
		ch >>= 1;
		i++;
	}
	
	// Stopbit: 1
    LED_OFF;
	waitBitTime();
}


void sendeText( char *text )
{
    uchar ch;
	
    do
	{
		ch = *text++;
		if ( ch != 0 )
		{
		    sendeZeichen( (uchar) ch ); 
		}
	}
	while ( ch != 0 );
}


void sendeDezimal( uint value )
{
	char buff[ 6 ];	// fr einen uint mit Wertebereich 0..65535 gengen 5 Zeichen plus Endenull
	
	// von hinten her die Ziffern berechnen und in den Buffer fllen
	char *ptr = &buff[ sizeof( buff ) ];
	*(--ptr) = 0;		// Endenull
	uchar first = 1;
	while ( ptr != buff )
	{
		// Divisionsrest durch 10 wird die (aktuell) niederwertigste Ziffer
		uchar rem = value % 10;
		value = value / 10;

		if ( ( rem == 0 ) && ( value == 0 ) && !first )
		{
			*(--ptr) = ' ';
		}
		else
		{
			*(--ptr) = rem + '0';
		}
		
		first = 0;
	}
	
	sendeText( buff );
}


void initHardware(void)
{
	// Clock-Prescaler auf 1 setzen -> maximale Geschwindigkeit
	CLKPR	= 0x80;	// erstmal freigeben
	CLKPR	= 0x00;	// dann setzen auf Faktor 1

	// Ports konfigurieren:
	// PB2  (5)		OUT OC0A	Motor/Servo rechts 
	// PB1  (3)		IN	Port	Hindernis-Taste links
	// PB0	(2)		IN	Port	Hindernis-Taste rechts / IR-Empfnger
	// PA7  (6) 	OUT	OC0B	Motor/Servo links
	// PA6  (7) 	OUT	OC1A	Lautsprecher
	// PA5  (8) 	-	-		-   (evtl. weiterer Servo)
	// PA4  (9) 	OUT	Port	LED rot
	// PA3  (10) 	OUT	Port	Trigger fr US-Sensor
	// PA2  (11) 	IN	Port	Signal von US-Sensor
	// PA1  (12) 	IN	ADC1	Sensor links
	// PA0  (13) 	IN	ADC0	Sensor rechts

	DDRA	= 0xD8;	// PORTA7 (u.a.) als Ausgang damit PWM funktioniert
	DDRB	= 0x04;	// PORTB2 als Ausgang damit PWM funktioniert
	PORTA	= 0x04;	// Pullup fr PA2 einschalten, falls nichts dranhngt
	PORTB	= 0x03;	// Pullup fr PB1 und PB0 einschalten

#ifdef SERVO_ANTRIEB
	// Timer0 auf Fast-PWM-Modus mit Teilerfaktor 64 setzen:
	TCCR0A	= 0xF3;	// Set on compare match (inverting), Fast-PWM-Mode
	TCCR0B	= 0x03;	// Prescaler 64
	OCR0A	= 0;	// erstmal aus
	OCR0B	= 0;	// erstmal aus
	TIMSK0	= 0x01;	// Interrupt freigeben: TOIE0 fr Timer0
#else
	// Timer0 auf Fast-PWM-Modus mit Teilerfaktor 1 setzen:
	TCCR0A	= 0xA3;	// Clear on compare match, Fast-PWM-Mode
	TCCR0B	= 0x01;	// Prescaler 1
	OCR0A	= 0;	// erstmal aus
	OCR0B	= 0;	// erstmal aus
	TIMSK0	= 0x01;	// Interrupt freigeben: TOIE0 fr Timer0
#endif
	
	// Timer1 auf CTC-Modus mit Waveform Generator auf Toggle
	TCCR1A	= 0x40; // toggle OC1A, CTC
	TCCR1B  = 0x18;	// TOP=ICR1, Clock stop
	

	// AD-Wandler initialisieren:
 	ADMUX	= 0x00;	// Vcc als Referenz, Mux auf ADC0
	ADCSRB	= 0x10;	// links ausgerichtet
	ADCSRA	= 0xCF;	// Wandler ein und starten, kein Auto-Trigger, Prescaler 128 => TODO: korrigieren: 208us pro AD-Wandlung => 624us pro Durchlauf

	adc_channel 	= 0;
	led_state 		= 0;
	timer_prescaler = 1;

	sei();			// Interrupts freigeben
}

